/*
  The l7-queue class handles libnetfilter-queue events and passes 
  packets to their appropriate conntack for classification. 
  
  By Ethan Sommer <sommere@users.sf.net> and Matthew Strait 
  <quadong@users.sf.net>, 2006-2007
  http://l7-filter.sf.net 

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version
  2 of the License, or (at your option) any later version.
  http://www.gnu.org/licenses/gpl.txt

  Based on nfqnl_test.c from libnetfilter-queue 0.0.12

  If you get error messages about running out of buffer space, increase it 
  with something like:

  echo 524280 > /proc/sys/net/core/rmem_default
  echo 524280 > /proc/sys/net/core/rmem_max
  echo 524280 > /proc/sys/net/core/wmem_default
  echo 524280 > /proc/sys/net/core/wmem_max
*/

using namespace std;
#include <pthread.h>

#include <iostream>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <map>
#include <netinet/in.h>

#include "l7-conntrack.h"
#include "l7-queue.h"
#include "util.h"

#include "queue-mt.h"


#define int_ntoa(x) inet_ntoa(*((struct in_addr *)&x))
#define MAX 64
#define MAX_PACKET_NUMBER 8    // Try to match up to 8 packets
#define SLEEP_MICRO_SECONDS 20 // When the queue is full or empty, just sleep a while
#define MAX_THREAD_NUMBER 7    // Maximum # of matching threads
#define QUEUE_SIZE 512


// Probably shouldn't really be global, but it's SO much easier
int maxpackets = 8; // by default.
int clobbermark = 0;

int total_num=0;
int num_thread=MAX_THREAD_NUMBER; // the actual input of threads. Default to MAX_THREAD_NUMBER.

extern unsigned int markmask;
extern unsigned int maskfirstbit;


extern "C" {
#include <linux/netfilter.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
}

#define NUM_THREADS 8
#define SLEEP_MICRO_SECONDS 20 // When the queue is full or empty, just sleep a while

typedef QMT<l7_connection *> conn_queue; // the connection queue type
conn_queue l7_queues[NUM_THREADS]; // the actual connection queues for each thread

pthread_t threads[NUM_THREADS]; // the thread pool

struct pthread_data
{
    int thread_num;
    conn_queue * p_l7_queue;
    long q_length; // for load balance
};

pthread_data thread_data[NUM_THREADS];


l7_queue::l7_queue(l7_conntrack *connection_tracker) 
{
  l7_connection_tracker = connection_tracker;
}

void l7_queue::kill_threads(int s)
{
    // was there a reason for this?  It just makes it crash on Ctrl-C for me.
    // delete l7_connection_tracker;
    l7printf(0, "Terminating process.\n");
    for (int i = 1; i <= num_thread; i++)
    {
        pthread_kill(threads[i], SIGKILL);
    }

    l7printf(0, "L7-filter process terminated.\n");

    exit(0);
}

static void * regex_matching_thread(void * data);

static int init_threads()
{
    int i = 0;
    int res = 0;

    for (i = 1; i <= num_thread; i++)
    {
        if (l7_queues[i].Init(1 << 16) == false)
            return 1;
        thread_data[i].thread_num = i;
        thread_data[i].p_l7_queue = &l7_queues[i];

        // Get timing information for each thread.
//        bzero((char *) &thread_tz, sizeof(thread_tz));
//        gettimeofday(&thread_start_time[i-1], &thread_tz);

        res = pthread_create(&threads[i], NULL, regex_matching_thread, 
				(void *) & thread_data[i]);
        if (res != 0)
            return res;
    }
    return res;
}


l7_queue::~l7_queue() 
{
}

struct nfq_q_handle *backup_qh; // for the connections in each thread
				// to set back the verdict for each packet.
void l7_queue::start(int queuenum) 
{
  struct nfq_handle *h;
  struct nfq_q_handle *qh;
  struct nfnl_handle *nh;
  int fd;
  int rv;
  char buf[4096];

  l7printf(3, "opening library handle\n");
  h = nfq_open();
  if(!h) {
    cerr << "error during nfq_open()\n";
    exit(1);
  }

  l7printf(3, "unbinding existing nf_queue handler for AF_INET (if any)\n");
  if(nfq_unbind_pf(h, AF_INET) < 0) {
    cerr << "error during nfq_unbind_pf()\n";
    exit(1);
  }

  l7printf(3, "binding nfnetlink_queue as nf_queue handler for AF_INET\n");
  if(nfq_bind_pf(h, AF_INET) < 0) {
    cerr << "error during nfq_bind_pf()\n";
    exit(1);
  }

  l7printf(3, "binding this socket to queue '0'\n");

// #######################
// register l7_queue_cb for the queue
// #######################
  qh = nfq_create_queue(h, queuenum, &l7_queue_cb, this);
  backup_qh = qh;
  if(!qh) {
    cerr << "error during nfq_create_queue()\n";
    exit(1);
  }

  l7printf(3, "setting copy_packet mode\n");
  if(nfq_set_mode(qh, NFQNL_COPY_PACKET, 0xffff) < 0) {
    cerr << "can't set packet_copy mode\n";
    exit(1);
  }

  nh = nfq_nfnlh(h);
  fd = nfnl_fd(nh);

  /* Create the threads to do regular expression there. */
  int res = init_threads();
  if (res)
  {
    l7printf(0, "Failed to creat the regular expression threads\n");
    exit(1);
  }

  // this is the main loop
  while (true){
  // #######################
  // This recv() is to receive from the queued packets
  // instead of packets from the NIC directly
  // #######################
    while ((rv = recv(fd, buf, sizeof(buf), 0)) && rv >= 0)
      nfq_handle_packet(h, buf, rv);
//total_num++;    
    cerr << "Error: recv() returned negative value." << endl;
    cerr << "rv=" << rv << endl;
    cerr << "errno=" << errno << endl;
    cerr << "errstr=" << strerror(errno) << endl << endl;
  }
  l7printf(3, "unbinding from queue 0\n");
  nfq_destroy_queue(qh);

  l7printf(3, "closing library handle\n");
  nfq_close(h);

  // Notify all subthreads to exit
  for (int i = 1; i <= num_thread; i++)
  {
      while(thread_data[i].p_l7_queue->Push(NULL) == false)
      {
          usleep(SLEEP_MICRO_SECONDS);
      }
      l7printf(0, "Waiting thread %d to finish.\n", i);
      pthread_join(threads[i], NULL);
    }
  exit(0);
}

u_int32_t l7_queue::handle_packet(nfq_data * tb, struct nfq_q_handle *qh) 
{
  int id = 0, ret, dataoffset, datalen;
  u_int32_t wholemark, mark, ifi; 
  struct nfqnl_msg_packet_hdr *ph;
  char * data;
  l7_connection * connection;

  total_num++; // accumulate the number of packets
  ph = nfq_get_msg_packet_hdr(tb);
  if(ph){
    id = ntohl(ph->packet_id);
    l7printf(4, "hw_protocol = 0x%04x hook = %u id = %u ", 
      ntohs(ph->hw_protocol), ph->hook, id);
  }

  // Need to get the wholemark so that we can pass the unmasked part back
  // Except for the print statement and debugging, there's not really any
  // reason to pull out the masked part, because it's always modified without
  // looking at it...
  wholemark = (nfq_get_nfmark(tb));
  if(clobbermark){
     mark = UNTOUCHED;
     wholemark = wholemark&(~markmask); // zero out our part of the mark
  }
  else mark = ((wholemark&markmask) >> maskfirstbit);
  l7printf(4, "wholemark = %#08x ", wholemark);
  l7printf(4, "mark = %d ", mark);

  ifi = nfq_get_indev(tb);
  if(ifi) l7printf(4, "indev = %d ", ifi);

  ifi = nfq_get_outdev(tb);
  if(ifi) l7printf(4, "outdev = %d ", ifi);

  ret = nfq_get_payload(tb, &data);
  if(ret >= 0) l7printf(4, "payload_len = %d\n", ret);
  
  char ip_protocol = data[9];

// ################################
// Why should we ingore non-TCP/UDP?
// Is it for packets?
// nfq_set_verdict is the write back function from 
// userspace to kernel netfilter QUEUE
// ################################
  // Ignore anything that's not TCP or UDP
  if(ip_protocol != IPPROTO_TCP && ip_protocol != IPPROTO_UDP)
    return nfq_set_verdict(qh, id, NF_ACCEPT, 0, NULL);

  dataoffset = app_data_offset((const unsigned char*)data);
  datalen = ret - dataoffset;

  //find the conntrack 
  string key = get_conntrack_key((const unsigned char*)data, false);
  connection = l7_connection_tracker->get_l7_connection(key);

  if(connection)
    l7printf(3, "Found connection orig:\t%s\n", key.c_str());

  if(!connection){
    //find the conntrack (backwards)
    string key = get_conntrack_key((const unsigned char*)data, true);
    connection = l7_connection_tracker->get_l7_connection(key);
  
    if(connection)
      l7printf(3, "Found connection reply:\t%s\n", key.c_str());
  
    // It seems to routinely not get the UDP conntrack until the 2nd or 3rd
    // packet.  Tested with DNS.
    if(!connection)
      l7printf(2, "Got packet, had no ct:\t%s\n", key.c_str());
  }

  // mark = the mark we found on the packet
  // connection->get_mark() = the mark that we have made internally
  if(connection){
    connection->increment_num_packets();
  
    if(datalen <= 0){
      l7printf(3, "Connection with no new application data ignored.\n");
      mark = NO_MATCH_YET; // no application data
    }
    else{
      if(connection->get_mark() != NO_MATCH_YET && 
         connection->get_mark() != UNTOUCHED){
        // It is classified already.  Reapply existing mark.
        mark = connection->get_mark();
      }
      else if(connection->get_num_packets() <= maxpackets){
        // Do the heavy lifting.
        connection->append_to_buffer((char*)(data+dataoffset),ret-dataoffset); 
        connection->IDs[connection->get_num_packets()-1] = id;
	connection->wholemarks[connection->get_num_packets()-1] = wholemark;
	l7printf(3, "Packet #%d, data is: %s\n", connection->get_num_packets(),
                 friendly_print((unsigned char *)connection->buffer,
                                connection->lengthsofar).c_str());
// ###################
// The matching thread!!!!!          
// Actually, it calls the classify function in l7-connection
// which in turn calls the classify function in l7-classify
// which calls regexec()
// But why do we need to decide if the connection has been matched in here???
// It has to be decided in l7-connection->classify() anyway.
// ###################
// #################################################################
// Instead of using the original single-threaded classify function
// We call "schedule" here to push the connection to a corresponding
// thread's runqueue
        schedule(connection);
// option to call HRW-based scheduler
//	HRW_schedule(connection);

	return 0; // packets' marks are set in each thread for all the packets
		  // in the same connection.
//	mark = connection->classify();
//	mark = connection->get_mark();
// #################################################################
/*        if(mark != NO_MATCH_YET){ // Got a match, no need to keep data
          free(connection->buffer);
          connection->buffer = NULL; // marks it not to be free'd again
        }
*/
      }
      else{ // num_packets > maxpackets and hasn't been classified
        mark = NO_MATCH;
        // if this is the first packet after we've given up, clean up
        if(connection->get_num_packets() == maxpackets+1){
          print_give_up(key, (unsigned char *)connection->buffer, 
                        connection->lengthsofar);
        
          free(connection->buffer);
          connection->buffer = NULL; // marks it not to be free'd again
        } // endif should clean up
      } // endif whether should run match or what
    } // endif there is any new data
  } // endif we found the connection
  else{
    l7printf(3, "Didn't yet find\t%s\n", key.c_str());
    mark = NO_MATCH_YET;
  }

  if(mark == UNTOUCHED) cerr << "NOT REACHED. mark is still UNTOUCHED.\n";

  l7printf(4, "Set verdict: ACCEPT %#08x\n", (mark<<maskfirstbit)|wholemark);
// ###############################
// Here comes the nfq_set_verdict_mark (refer to line: 181)
// ###############################
  return nfq_set_verdict_mark(qh, id, NF_ACCEPT, 
                              htonl((mark<<maskfirstbit)|wholemark), 0, NULL);
}

// Returns a string that uniquely defines the connection
string l7_queue::get_conntrack_key(const unsigned char *data, bool reverse) 
{
  char * buf = (char *)malloc(256);
  int ip_hl = 4*(data[0] & 0x0f);
  char ip_protocol = data[9];

  if(ip_protocol == IPPROTO_TCP){
    if(reverse){
      snprintf(buf, 255, 
              "tcp      6 src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d",
	      data[12], data[13], data[14], data[15],
	      data[16], data[17], data[18], data[19],
	      data[ip_hl]*256+data[ip_hl+1], data[ip_hl+2]*256+data[ip_hl+3]);
    }
    else{
      snprintf(buf, 255, 
              "tcp      6 src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d",
	      data[16], data[17], data[18], data[19],
	      data[12], data[13], data[14], data[15],
	      data[ip_hl+2]*256+data[ip_hl+3], data[ip_hl]*256+data[ip_hl+1]);
    }
  }
  else if(ip_protocol == IPPROTO_UDP){
    if(reverse){
      snprintf(buf, 255, 
              "udp      17 src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d",
	      data[12], data[13], data[14], data[15],
	      data[16], data[17], data[18], data[19],
	      data[ip_hl]*256+data[ip_hl+1], data[ip_hl+2]*256+data[ip_hl+3]);
    }
    else{
      snprintf(buf, 255, 
              "udp      17 src=%d.%d.%d.%d dst=%d.%d.%d.%d sport=%d dport=%d",
	      data[16], data[17], data[18], data[19],
	      data[12], data[13], data[14], data[15],
	      data[ip_hl+2]*256+data[ip_hl+3], data[ip_hl]*256+data[ip_hl+1]);
    }
  }
  else{
    // ######################
    // What if it's some other protocols?
    // unsupported??? Refers to line:202
    // ######################
    l7printf(0, "Tried to get conntrack key for unsupported protocol!\n");
    buf[0] = '\0';
  }
  string answer = buf;
  free(buf);

  l7printf(3, "Made key from packet:\t%s\n", answer.c_str());

  return answer;
}

/* Returns offset the into the skb->data that the application data starts */
int l7_queue::app_data_offset(const unsigned char *data)
{
  int ip_hl = 4*(data[0] & 0x0f);
  char ip_protocol = data[9];

  if(ip_protocol == IPPROTO_TCP){
    // 12 == offset into TCP header for the header length field.
    int tcp_hl = 4*(data[ip_hl + 12]>>4);
    return ip_hl + tcp_hl;
  }
  else if(ip_protocol == IPPROTO_UDP){
    return ip_hl + 8; /* UDP header is always 8 bytes */
  }
  else{
      l7printf(0, "Tried to get app data offset for unsupported protocol!\n");
      return ip_hl + 8; /* something reasonable */
  }
}

static int l7_queue_cb(struct nfq_q_handle *qh, struct nfgenmsg *nfmsg,
		       struct nfq_data *nfa, void *data) 
{
  struct nfqnl_msg_packet_hdr *ph;

  u_int32_t id = 0;
  ph = nfq_get_msg_packet_hdr(nfa);
  if(ph)
    id = ntohl(ph->packet_id);
  
  u_int32_t wholemark = nfq_get_nfmark(nfa);

  // If it already has a mark (and we don't want to clobber it), 
  // just pass it back with the same mark
  if((wholemark<<maskfirstbit)&markmask != UNTOUCHED && !clobbermark){
    static unsigned int naaltered = 0;
    naaltered++;
    if((naaltered^(naaltered-1)) == (2*naaltered-1)) // is it a power of 2?
      cerr << "My part of the mark has already been altered, ignoring these "
              "packets!\n(" << naaltered << " ignored so far.) "
              "Fix your rules or use l7-filter -c.\n";
// ###########################
// nfq_set_verdict_mark AGAIN!!!
// ###########################
    return nfq_set_verdict_mark(qh, id, NF_ACCEPT, htonl(wholemark), 0, NULL);
  }
// ###########################
// The l7-queue callback function actually calls handle_packet()
// ###########################
  return ((l7_queue *)data)->handle_packet(nfa, qh);
}

void l7_queue::schedule(l7_connection * p_l7_conn)
{
    // Here it is round robin like policy 
    // to distribute the packets to all available 
    // regular expression engines
/*
    while (1) {
        rr_pointer++;
        if (rr_pointer > num_thread)
            rr_pointer=1;
        if (thread_data[rr_pointer].p_l7_queue->Push(p_l7_conn) == true)
        {
            p_l7_conn->set_thread_num(rr_pointer);
            break;
        }
    }
*/

/************************************************************************/
    // Comment out this section if you want to try out the HRW-based scheduler
    // Here is an connection-based scheduler to 
    // more efficiently utilize the multi-core
    // new connection
    if (p_l7_conn->get_thread_num() == 0)
    {
        while (1)
        {
            // This is a load balance to find 
	    // the most idling thread for this new connection
            int least=1;
            for (int i=1; i<=num_thread; i++)
            {
                if (thread_data[least].p_l7_queue->GetEntryNum() > 
			thread_data[i].p_l7_queue->GetEntryNum())
                    least = i;
            }
            if (thread_data[least].p_l7_queue->Push(p_l7_conn) == true)
            {
                p_l7_conn->set_thread_num(least);
                break;
            }
        }
    }
    else
    {
        while (thread_data[p_l7_conn->get_thread_num()].p_l7_queue->Push(p_l7_conn) == false)
            usleep(SLEEP_MICRO_SECONDS);
    }
}

/*******************************************************************
        To computer the theory adjust value based on Ross's paper.
int HRW(int conn_id, unsigned int buff_len)
{
        // Schedule to a Core first
        for (int i=0; i<8; i++)
        {
                // find the minimum core_len
                if (core_len[core_len_min_id] > core_len[i])
                        core_len_min_id = i;
        }
        // recursive calculation
        double p[8];// sum(p_i)=1; it's a influence ratio
        double sum=0.0;
        for (int i=0; i<8; ++i) {
                sum+=(core_len[core_len_min_id]/core_len[i]);
        }
        for (int i=0; i<8; ++i) {
                p[i] = (core_len[core_len_min_id]/core_len[i])/sum;
        }
        double x[8]; // x[i] is the theoretically correct adjustement for PU[i]
        x[0]=pow(8*p[0], 1/8);
        for (int i=1; i<8; ++i) {
                double mul_xi=1.0;
                for (int j=0; j<i; ++j)
                        mul_xi*=x[j];
                x[i]=pow(pow(x[i-1], 8-i)+(N-i)*(p[i]-p[i-1])/mul_xi, 1/(8-i));
        }
        // apply x[0..7] to all the cores
        for (int i=0; i<8; i++)
        {
                core_weights[i] = x[i]* abs((signed int)(1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (core_weights[core_weight_max_id] < core_weights[i])
                        core_weight_max_id = i;
        }
        core_select = core_weight_max_id;
        core_len[core_select]+=(int)buff_len;

        // Schedule to a Pipe on core_select
        for (int i=0; i<2; i++)
        {
                // find the minimum pipe_len
                if (pipe_len[core_select][pipe_len_min_id] > pipe_len[core_select][i])
                        pipe_len_min_id = i;
        }
        // recursive calculation
        sum=0.0; // reset sum
        for (int i=0; i<2; ++i) {
                sum+=(pipe_len[pipe_len_min_id]/pipe_len[i]);
        }
        for (int i=0; i<2; ++i) {
                p[i] = (pipe_len[pipe_len_min_id]/pipe_len[i])/sum;
        }
        x[0]=pow(2*p[0], 1/2);
        for (int i=1; i<2; ++i) {
                double mul_xi=1.0;
                for (int j=0; j<i; ++j)
                        mul_xi*=x[j];
                x[i]=pow(pow(x[i-1], 2-i)+(N-i)*(p[i]-p[i-1])/mul_xi, 1/(2-i));
        }
        // apply x[0..1] to all the cores

        for (int i=0; i<2; i++)
        {
                pipe_weights[core_select][i] = x[i] * abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (pipe_weights[core_select][pipe_weight_max_id] < pipe_weights[core_select][i])
                        pipe_weight_max_id = i;
        }
        pipe_select = pipe_weight_max_id;
        pipe_len[core_select][pipe_select]+= (int)buff_len;

        // Schedule to a thread on pipe_select on core_select
        for (int i=0; i<4; i++)
        {
                // find the minimum thread_len
                if (thread_len[core_select][pipe_select][thread_len_min_id] > thread_len[core_select][pipe_select][i])
                        thread_len_min_id = i;
        }
        // recursive calculation
        sum=0.0; // reset sum
        for (int i=0; i<4; ++i) {
                sum+=(thread_len[thread_len_min_id]/thread_len[i]);
        }
        for (int i=0; i<4; ++i) {
                p[i] = (thread_len[thread_len_min_id]/thread_len[i])/sum;
        }
        x[0]=pow(4*p[0], 1/4);
        for (int i=1; i<4; ++i) {
                double mul_xi=1.0;
                for (int j=0; j<i; ++j)
                        mul_xi*=x[j];
                x[i]=pow(pow(x[i-1], 4-i)+(N-i)*(p[i]-p[i-1])/mul_xi, 1/(4-i));
        }
        // apply x[0..3] to all the cores

        for (int i=0; i<4; i++)
        {
                thread_weights[core_select][pipe_select][i] = x[i] * abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (thread_weights[core_select][pipe_select][thread_weight_max_id] < thread_weights[core_select][pipe_select][i])
                        thread_weight_max_id = i;
        }
        thread_select = thread_weight_max_id;
        thread_len[core_select][pipe_select][thread_select]+= (int)buff_len;
//cout<<"\tthread select ="<<thread_select<<endl;
        return core_select*8 + pipe_select*4 + thread_select;

}
********************************************/

/***********************************************************************/
/** Below is the HRW-based scheduler
    Uncomment to try it out
    Note that it is currently designed to handle the Sun Niagara 2 multicore architecture,
    where there are 8 cores, each holding 2 pipelines with 4 HW threads in each pipeline.
    You will need to adjust the corresponding array size to your specific architecture.
**/
/**
int MotherMatrix[8][8]={1,1.5,1.5,1.5,3,3,3,3,1.5,1,1.5,1.5,3,3,3,3,1.5,1.5,1,1.5,3,3,3,3,1.5,1.5,1.5,1,3,3,3,3,3,3,3,3,1,1.5,1.5,1.5,3,3,3,3,1.5,1,1.5,1.5,3,3,3,3,1.5,1.5,1,1.5,3,3,3,3,1.5,1.5,1.5,1}; // for cache-aware AHRW (CA-AHRW). Define your own communication matrix and use it as a multiplier to the weight generator

signed long core_weights[8] = {0, 0, 0, 0, 0, 0, 0, 0};
signed long pipe_weights[8][2]= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
signed long thread_weights[8][2][4] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int core_len[8] = {0, 0, 0, 0, 0, 0, 0, 0};
int pipe_len[8][2]={0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int thread_len[8][2][4] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int core_weight_max_id = 0;
int core_len_min_id =0;
#define CORE_LEN_THRESHOLD 100 // usually packets are 200 bytes on average,
                               // and to classify a conn need less than 2 packets
int core_select =0;
int pipe_weight_max_id = 0;
int pipe_len_min_id =0;
#define PIPE_LEN_THRESHOLD 100 // usually packets are 200 bytes on average,
                               // and to classify a conn need less than 2 packets
int pipe_select = 0;
int thread_weight_max_id = 0;
int thread_len_min_id =0;
#define THREAD_LEN_THRESHOLD 100 // usually packets are 200 bytes on average,
                               // and to classify a conn need less than 2 packets
int thread_select = 0;
#include <math.h>
// Now it's the Adjusted Hightest Random Weight (AHRW) based scheduling
int HRW(int conn_id, unsigned int buff_len)
{
        // Schedule to a Core first
        for (int i=0; i<8; i++)
        {
                // find the minimum core_len
                if (core_len[core_len_min_id] > core_len[i])
                        core_len_min_id = i;
        }
        for (int i=0; i<8; i++)
        {
                double adjust = 0.0;
                if (core_len[core_len_min_id] != 0 && core_len[i]!=0)
                        adjust = core_len[core_len_min_id] / core_len[i];
                else
                        adjust = 1;
                core_weights[i] = adjust * abs((signed int)(1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (core_weights[core_weight_max_id] < core_weights[i])
                        core_weight_max_id = i;
        }
        core_select = core_weight_max_id;
        core_len[core_select]+=(int)buff_len;
//cout<<"core select ="<<core_select<<endl;

        // Schedule to a Pipe on core_select
        for (int i=0; i<2; i++)
        {
                // find the minimum pipe_len
                if (pipe_len[core_select][pipe_len_min_id] > pipe_len[core_select][i])
                        pipe_len_min_id = i;
        }
        for (int i=0; i<2; i++)
        {
                double adjust = 0.0;
                if (pipe_len[core_select][pipe_len_min_id] != 0 && pipe_len[core_select][i] != 0)
                        adjust = pipe_len[core_select][pipe_len_min_id] / pipe_len[core_select][i];
                else
                        adjust = 1;
                pipe_weights[core_select][i] = adjust * abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (pipe_weights[core_select][pipe_weight_max_id] < pipe_weights[core_select][i])
                        pipe_weight_max_id = i;
        }
        pipe_select = pipe_weight_max_id;
        pipe_len[core_select][pipe_select]+= (int)buff_len;
//cout<<"\tpipe select ="<<pipe_select<<endl;

        // Schedule to a thread on pipe_select on core_select
        for (int i=0; i<4; i++)
        {
                // find the minimum thread_len
                if (thread_len[core_select][pipe_select][thread_len_min_id] > thread_len[core_select][pipe_select][i])
                        thread_len_min_id = i;
        }
        for (int i=0; i<4; i++)
        {
                double adjust = 0.0;
                if (thread_len[core_select][pipe_select][thread_len_min_id] != 0 && thread_len[core_select][pipe_select][i] != 0)
                        adjust = thread_len[core_select][pipe_select][thread_len_min_id] / thread_len[core_select][pipe_select][i];
                else
                        adjust = 1;
                thread_weights[core_select][pipe_select][i] = adjust * abs((1103515245*((1103515245*i+12345)^conn_id)+12345)%((signed int)(pow(2, 31))));
                if (thread_weights[core_select][pipe_select][thread_weight_max_id] < thread_weights[core_select][pipe_select][i])
                        thread_weight_max_id = i;
        }
        thread_select = thread_weight_max_id;
        thread_len[core_select][pipe_select][thread_select]+= (int)buff_len;
//cout<<"\tthread select ="<<thread_select<<endl;
        return core_select*8 + pipe_select*4 + thread_select;

}

// the HRW scheduler
void HRW_schedule(l7_connection * p_l7_conn)
{
        pthread_mutex_lock( &mutex1 );
        int thread_id = HRW(p_l7_conn->id, p_l7_conn->get_buffer_cur_length());
        thread_data[thread_id].p_l7_queue->Push(p_l7_conn);
        pthread_mutex_unlock( &mutex1 );
}

************************************************************************************/

int n_connection=0;
// Mark the overall queue length per thread
long ql[MAX_THREAD_NUMBER];

static void * regex_matching_thread(void * data)
{
    pthread_data * pdata = (struct pthread_data *)(data);

    /*Now start to the regular experession work*/
    l7printf(0, "Thread %d starts to work right now.\n", pdata->thread_num);
/********************************************************************/
    // comment out this section if you want to try out the HRW-based scheduler

    // set processor affinity
    // each connection take over 1 processor.

    cpu_set_t mask;
    CPU_ZERO(&mask);
    CPU_SET(pdata->thread_num-1, &mask);
    unsigned int len = sizeof(mask);
    if (pthread_setaffinity_np(pthread_self(), len, &mask) < 0) {
        perror("pthread_setaffinity_np");
    }

/********************************************************************/
    while (1)
    {
        l7_connection * p_l7_connection = NULL;
        while(pdata->p_l7_queue->Pop(p_l7_connection) == false)
            usleep(SLEEP_MICRO_SECONDS);
        if (p_l7_connection == NULL)
            break;
        unsigned int mark = p_l7_connection->get_mark();
    
	// Check again to see if the connection has been classified by previous
	// matching. This is possible because while conn#c+1 is Enqueued,
	// conn#c can still be in the process of classification.
	if (mark != UNTOUCHED && mark != NO_MATCH_YET && mark != NO_MATCH)
            continue;

//        p_l7_connection->reassembly_new_packet();
//        char *buffer = p_l7_connection->get_buffer();

/***************************************************************
     For HRW to maintain load balance
     accumulate the thread queue length
int cur_len = p_l7_connection->get_buffer_cur_length();
pdata->q_length += cur_len;
***************************************************************/

        mark = p_l7_connection->classify();
// directly copy from handle_packet
        if(mark != NO_MATCH_YET){ // Got a match, no need to keep data
          free(p_l7_connection->buffer);
          p_l7_connection->buffer = NULL; // marks it not to be free'd again
        }

        if(mark == UNTOUCHED) cerr << "NOT REACHED. mark is still UNTOUCHED.\n";
	
// ###############################
// Here comes the nfq_set_verdict_mark (refer to line: 181)
// ###############################
	for (int i=0; i<p_l7_connection->get_num_packets(); i++)
	{
	        l7printf(4, "Set verdict: ACCEPT %#08x\n",
	                (mark<<maskfirstbit)|p_l7_connection->wholemarks[i]);
		nfq_set_verdict_mark(backup_qh, p_l7_connection->IDs[i], NF_ACCEPT,
                      htonl((mark<<maskfirstbit)|p_l7_connection->wholemarks[i]), 
			0, NULL);
	}
//        p_l7_connection->set_mark(mark);
//        n_connection++;
/*
        if (mark != UNTOUCHED && mark != NO_MATCH_YET && mark != NO_MATCH)
        {
            if (p_l7_connection->get_packet_number() < MAX_PACKET_NUMBER )
                fprintf(output2, "%d\n", p_l7_connection->get_buffer_cur_length());
            l7printf(1, "Match %s:\t%s by thread %d.\n",
            l7_classifier->get_pattern_name(mark).c_str(),
            p_l7_connection->get_key().c_str(),
            pdata->thread_num);
        }
*/

/***********************************************************************
	Again, this section is for HRW-based scheduler to maintain queue length
	Reduce the queue size after the connection buffer is processed.
        int c_id = pdata->thread_num/8;
        int p_id = (pdata->thread_num%8)/4;
        int t_id = (pdata->thread_num%8)%4;
        thread_len[c_id][p_id][t_id] -= cur_len;
        pipe_len[c_id][p_id] -= cur_len;
        core_len[c_id] -= cur_len;
************************************************************************/
    }


    l7printf(0, "Thread %d finished.\n", pdata->thread_num);

    // Get thread ending information
/*    bzero((char *) &thread_tz, sizeof(thread_tz));
    gettimeofday(&thread_end_time[pdata->thread_num-1], &thread_tz);
    TIMERSUB(&thread_end_time[pdata->thread_num-1],
                &thread_start_time[pdata->thread_num-1],
                &thread_diff_time[pdata->thread_num-1]);
    l7printf(0, "Running time for thread %d was %lu.%lu seconds\n",
                pdata->thread_num,
                (unsigned long)thread_diff_time[pdata->thread_num-1].tv_sec,
                (unsigned long)thread_diff_time[pdata->thread_num-1].tv_usec);
*/

    // calculate the overall thread queue length
    ql[pdata->thread_num] = pdata->q_length;

    pthread_exit(NULL);
}

